unit Http;

interface

uses NSys, NLib, NUtil, NSock; 

type
  TResponse = record
    Status: Integer;
    ReasonPhrase: string;
    LastModified: string;
    Body: string;
  end;

  THttp = class(TObject)
  private
    FBuf: string;
    FHost: string;
    FIMS: string;
    FPath: string;
    FPort: Integer;
    FTcp: TTcp;
    FUrl: string;
    function ParseResponse: TResponse;
    procedure TcpGetHostByName(Sender: TObject; IpAddr: string);
    procedure TcpConnect(Sender: TObject);
    procedure TcpRecv(Sender: TObject);
    procedure TcpClose(Sender: TObject);
    procedure TcpError(Sender: TObject; ErrorCode: Integer);
  public
    UserAgent: string;
    Getting: Boolean;
    OnError: procedure (ErrorCode: Integer) of object;
    OnGet: procedure (const Response: TResponse) of object;
    OnMsg: procedure (const Msg: string) of object;
    constructor Create;
    destructor Destroy; override;
    procedure Get(const Url: string; const IMS: string);
  end;

implementation

procedure UrlSplit(const S: string; var Scheme, NetLoc, Path: string);
var
  i: Integer;
  tmp: string;
begin
  i := 0;
  Scheme := '';
  NetLoc := '';
  Path := '';
  while i < Length(S) do
  begin
    Inc(i);
    case S[i] of
      ':':
        if Scheme = '' then
        begin
          Scheme := tmp;
          tmp := '';
        end else
          tmp := tmp + S[i];
      '/':
        if (NetLoc = '') and (tmp <> '/') then
        begin
          NetLoc := tmp;
          tmp := '';
        end else
          tmp := tmp + S[i];
    else
      tmp := tmp + S[i];
    end;
  end;
  Path := '/' + tmp;
end;

constructor THttp.Create;
begin
  FTcp := TTcp.Create;
  FTcp.OnGetHostByName := TcpGetHostByName;
  FTcp.OnConnect := TcpConnect;
  FTcp.OnRecv := TcpRecv;
  FTcp.OnClose := TcpClose;
  FTcp.OnError := TcpError;
end;

destructor THttp.Destroy;
begin
  FTcp.Free;
end;

procedure THttp.Get(const Url: string; const IMS: string);
var
  i: Integer;
  s: string;
  scheme, loc: string;
begin
  if Getting then
    Exit;

  Getting := True;
  FHost := '';
  FBuf := '';
  FUrl := Url;
  FIMS := IMS;
  UrlSplit(FUrl, scheme, loc, FPath);
  for i := 1 to Length(loc) do
    if (loc[i] = ':') and (FHost = '') then
    begin
      FHost := s;
      s := '';
    end else
      s := s + loc[i];

  if FHost = '' then
  begin
    FHost := loc;
    FPort := 80;
  end else
    FPort := SInt(s);

  FTcp.GetHostByName(FHost);
end;

function THttp.ParseResponse;
  procedure ParseStatusLine(const S: string; var Code: Integer;
    var Phrase: string);
  var
    i, len: Integer;
  begin
    Code := 0;
    len := Length(S);
    if len > 0 then
    begin
      if S[len] = #10 then
        Dec(len);
      if S[len] = #13 then
        Dec(len);
    end;
    for i := 1 to len do
      if S[i] = ' ' then
      begin
        Code := SInt(Copy(S, i+1, 3));
        Phrase := Copy(S, i+5, len-(i+5)+1);
        Exit;
      end;
  end;
  procedure ParseHeader(const S: string; var Name, Value: string);
  var
    i, len: Integer;
  begin
    len := Length(S);
    if len > 0 then
    begin
      if S[len] = #10 then
        Dec(len);
      if S[len] = #13 then
        Dec(len);
    end;
    for i := 1 to len do
      if S[i] = ':' then
      begin
        Name := Copy(S, 1, i-1);
        Value := Copy(S, i+2, len-(i+2)+1);
        Exit;
      end;
    Name := '';
    Value := '';
  end;
var
  eoh: Boolean;
  s, n, v: string;
begin
  Result.Status := 0;
  Result.Body := '';
  Result.LastModified := '';
  eoh := False;
  with TSplitLines.Create(FBuf) do
    try
      while GetLine(s) do
        if Result.Status = 0 then
          ParseStatusLine(s, Result.Status, Result.ReasonPhrase)
        else if eoh then
          Result.Body := Result.Body + s
        else
          if s = #13#10 then
            eoh := True
          else begin
            ParseHeader(s, n, v);
            if Lower(n) = 'last-modified' then
              Result.LastModified := v;
          end;
    finally
      Free;
    end;
end;

procedure THttp.TcpGetHostByName;
begin
  OnMsg(Fmt('Connecting to %s...', [FHost]));
  (Sender as TTcp).Connect(IpAddr, FPort);
end;

procedure THttp.TcpConnect(Sender: TObject);
var
  header: string;
begin
  OnMsg(Fmt('Connected. Sending request to %s...', [FHost]));
  header := Fmt('Host: %s'#13#10, [FHost]);
  if FIMS <> '' then
    header := header + Fmt('If-Modified-Since: %s'#13#10, [FIMS]);
  if UserAgent <> '' then
    header := header + Fmt('User-Agent: %s'#13#10, [UserAgent]);
  (Sender as TTcp).Send(Fmt('GET %s HTTP/1.0'#13#10'%s'#13#10,
    [FPath, Header]));
end;

procedure THttp.TcpRecv(Sender: TObject);
var
  s: string;
begin
  repeat
    s := (Sender as TTcp).Recv(4096);
    FBuf := FBuf + s; // XXX
  until s = '';
end;

procedure THttp.TcpClose(Sender: TObject);
begin
  Getting := False;
  OnGet(ParseResponse);
end;

procedure THttp.TcpError(Sender: TObject; ErrorCode: Integer);
begin
  Getting := False;
  OnError(ErrorCode);
end;

end.
