unit Sntp;

interface

uses NSock;

type
  TNtpTime = packed record
    Seconds: Cardinal;
    Fraction: Cardinal;
  end;

  PNtpPacket = ^TNtpPacket;
  TNtpPacket = packed record
    LI_VN_Mode: Byte; // li:2 vn:3 mode:3
    Stratum: Byte;
    Poll: Byte;
    Precision: Shortint;
    RootDelay: Integer;
    RootDispersion: Integer;
    RefId: Integer;
    RefTime: TNtpTime;
    Originate: TNtpTime;
    Receive: TNtpTime;
    Transmit: TNtpTime;
  end;

  TSntp = class(TObject)
  private
    FAdjustOffset: Double;
    FIpAddr: string;
    FOriginate: Double;
    FPort: Integer;
    FUdp: TUdp;
    procedure SendPacket;
    procedure UdpError(Sender: TObject; ErrorCode: Integer);
    procedure UdpGetHostByName(Sender: TObject; IpAddr: string);
    procedure UdpRecv(Sender: TObject);
  public
    Syncing: Boolean;
    OnError: procedure (ErrorCode: Integer; InvalidPacket: Boolean) of object;
    OnSync: procedure (const IpAddr: string; Stratum: Integer;
      Offset, Delay: Double; Adjusted: Boolean) of object;
    destructor Destroy; override;
    procedure Sync(const Server: string; Port: Integer; AdjustOffset: Double);
    procedure Abort;
  end;

implementation

uses Windows, WinSock;

function GetSysTime: Double;
var
  st: TSystemTime;
  days: Integer;
begin
  GetSystemTime(st);
  with st do
  begin
    if wMonth <= 2 then
    begin
      Inc(wMonth, 12);
      Dec(wYear);
    end;
    // days from 1900/01/01
    days := (wYear*365)+(wYear div 4)-(wYear div 100)+(wYear div 400) +
      (wMonth*30) + (((wMonth+1)*3) div 5) + wDay - 33 - 693960;
    Result := ((days-1) * 86400.0 + wHour * 3600.0 + wMinute * 60.0 + wSecond)+
      wMilliseconds / 1000.0;
  end;
end;

procedure SetSysTime(Value: Double);
var
  ft: TFileTime;
  st: TSystemTime;
begin
  Int64(ft) := 94354848000000000 + Trunc(Value * 10000000);
  FileTimeToSystemTime(ft, st);
  SetSystemTime(st);
end;

function DoubleTime(const T: TNtpTime): Double;
begin
  Result := Cardinal(ntohl(T.Seconds)) +
    Cardinal(ntohl(T.Fraction)) / ($FFFFFFFF + 1.0);
end;

function ValidatePacket(const pkt: TNtpPacket): Boolean;
begin
  Result := True;
  with pkt do
  begin
    if (LI_VN_Mode shr 6) > 2 then // li
      Result := False
    else if (LI_VN_Mode shr 3 and 7) <> 4 then // vn
      Result := False
    else if (LI_VN_Mode and 7) <> 4 then // mode
      Result := False
    else if Stratum > 14 then
      Result := False
    else if (Transmit.Seconds = 0) and (Transmit.Fraction = 0) then
      Result := False;
  end;
end;

destructor TSntp.Destroy;
begin
  FUdp.Free;
end;

procedure TSntp.Sync;
begin
  if Syncing then
    Exit;

  Syncing := True;
  FPort := Port;
  FAdjustOffset := AdjustOffset;
  if Assigned(FUdp) then
    FUdp.Free;
  FUdp := TUdp.Create;
  FUdp.OnGetHostByName := UdpGetHostByName;
  FUdp.OnRecv := UdpRecv;
  FUdp.OnError := UdpError;
  FUdp.GetHostByName(Server);
end;

procedure TSntp.Abort;
begin
  if not Syncing then
    Exit;
  FUdp.Free;
  FUdp := nil;
  Syncing := False;
end;

procedure TSntp.SendPacket;
var
  buf: string;
begin
  SetString(buf, nil, SizeOf(TNtpPacket));
  FillChar(buf[1], SizeOf(TNtpPacket), 0);
  PNtpPacket(@buf[1]).LI_VN_Mode := $23; // vn=4, mode=3 (client)
  FOriginate := GetSysTime; 
  FUdp.Send(buf, FIpAddr, FPort);
end;

procedure TSntp.UdpError;
begin
  OnError(ErrorCode, False);
end;

procedure TSntp.UdpGetHostByName;
begin
  FIpAddr := IpAddr;
  SendPacket; 
end;

procedure TSntp.UdpRecv;
var
  pkt: PNtpPacket;
  dest, recv, xmit, delay, ofs: Double;
  buf: string;
  adjusted: Boolean;
begin
  dest := GetSysTime;
  buf := FUdp.Recv(SizeOf(TNtpPacket));
  FUdp.Free;
  FUdp := nil;

  if Length(buf) <> SizeOf(TNtpPacket) then
  begin
    Syncing := False;
    OnError(0, True);
    Exit;
  end;

  pkt := PNtpPacket(@buf[1]);
  if not ValidatePacket(pkt^) then
  begin
    Syncing := False;
    OnError(0, True);
    Exit;
  end;

  recv := DoubleTime(pkt.Receive);
  xmit := DoubleTime(pkt.Transmit);
  delay := (dest - FOriginate) - (recv - xmit);
  ofs := ((recv - FOriginate) + (xmit - dest)) / 2;
  adjusted := (ofs > FAdjustOffset) or (ofs < -FAdjustOffset);
  if adjusted then
    SetSysTime(GetSysTime + ofs);

  Syncing := False;
  OnSync(FIpAddr, pkt.stratum, ofs, delay, adjusted);
end;

end.
